#!/usr/bin/env python
"""
A set of utilities to handle probability distributions
"""
from __future__ import annotations

import copy
import numbers
import random

import numpy as np
import scipy.stats as stats

from negmas.common import Distribution

__all__ = [
    "Distribution",  # THe interface of a distribution class
    "ScipyDistribution",  # A probability distribution using scipy stats
    "NormalDistribution",  # A probability distribution using scipy stats
    "UniformDistribution",  # A probability distribution using scipy stats
    "Real",
    "make_distribution",
]

EPSILON = 1e-8


class Real(Distribution):
    """A real number implementing the `Distribution` interface"""

    def __init__(
        self,
        loc: int | float | numbers.Real = 0.0,
        type: str = "",
        scale: float = 0.0,
        **kwargs,
    ):
        loc = float(loc)
        if not (0 <= scale <= EPSILON):
            raise ValueError(
                f"Cannot construct a Real with this scale ({scale}) It must be zero"
            )
        self._loc: float = loc
        self._scale = 0.0
        self._type = type if type else "uniform"

    def __float__(self) -> float:
        return self.loc

    def __call__(self, val: float) -> float:
        return 1.0 if abs(val - self.loc) < 1e-10 else 0.0

    @property
    def loc(self) -> float:
        """Returns the location of the distributon (usually mean)"""
        return self._loc

    def __copy__(self):
        """Copies the distribution"""
        return Real(type="", loc=self.loc)

    def __deepcopy__(self, memo):
        """Performs deep copying"""
        return Real(loc=self.loc)
        # todo: check that this works. I think it does not
        # super().__init__(type=type if type else "uniform")

    def type(self) -> str:
        return self._type

    def mean(self) -> float:
        """Finds the mean"""
        return self._loc

    def prob(self, val: float) -> float:
        """Returns the probability for the given value"""
        return 1.0 if self == val < EPSILON else 0.0

    def cum_prob(self, mn: float, mx: float) -> float:
        """Returns the probability for the given range"""
        return int(mn <= self.loc <= mx)

    def sample(self, size: int = 1) -> np.ndarray:
        """Samples `size` elements from the distribution"""
        return self._loc * np.ones(size)

    @property
    def scale(self) -> float:
        """Returns the scale of the distribution (may be std. dev.)"""
        return 0.0

    @property
    def min(self) -> float:
        """Returns the minimum"""
        return self._loc

    @property
    def max(self) -> float:
        """Returns the maximum"""
        return self._loc

    def is_gaussian(self):
        return True

    def is_uniform(self):
        return True

    def is_crisp(self) -> bool:
        """Returns true if this is a distribution with all probability at one point (delta(v))"""
        return self.scale < EPSILON / 1000

    def __add__(self, other):
        """Returns the distribution for the sum of samples of `self` and `other`"""
        if isinstance(other, float):
            return Real(float(self) + other)
        return other.__class__(loc=other.loc + self._loc, scale=other.scale)

    def __sub__(self, other):
        """Returns the distribution for the difference between samples of `self` and `other`"""
        if isinstance(other, float):
            return Real(float(self) - other)
        return other.__class__(loc=self.loc - other.loc, scale=other.scale)

    def __mul__(self, other):
        """Returns the distribution for the sum of samples of `self` and `other`"""
        if isinstance(other, float):
            return self._loc * other
        return other * self._loc

    def __lt__(self, other):
        """Check that a sample from `self` is ALWAYS less than a sample from other `other`"""
        if isinstance(other, float):
            return float(self) < other
        return self.max < other.min

    def __le__(self, other):
        """Check that a sample from `self` is ALWAYS less or equal a sample from other `other`"""
        if isinstance(other, float):
            return float(self) <= other
        return self.max <= other.min

    def __eq__(self, other):
        """Checks for equality of the two distributions"""
        if isinstance(other, Distribution):
            return other.scale < EPSILON and abs(other.loc - self.loc) < EPSILON
        if isinstance(other, float):
            return abs(self.loc - other) < EPSILON
        raise ValueError(f"Cannot compare CrispValues with {type(other)}")

    def __ne__(self, other):
        """Checks for ineqlaity of the distributions"""
        return not self == other

    def __gt__(self, other):
        """Check that a sample from `self` is ALWAYS greater than a sample from other `other`"""
        if isinstance(other, float):
            return float(self) > other
        return self.min > other.max

    def __ge__(self, other):
        """Check that a sample from `self` is ALWAYS greater or equal a sample from other `other`"""
        if isinstance(other, float):
            return float(self) >= other
        return self.min >= other.max

    def __str__(self):
        """Converts to a readable string"""
        return str(float(self._loc))

    __repr__ = __str__


def make_distribution(x: int | float | numbers.Real | Distribution) -> Distribution:
    """
    Ensures the output is `Distribution`

    Remarks:
        The input can either be `Distribution` or a number

    """
    if not isinstance(x, Distribution):
        return Real(x)
    return x


class ScipyDistribution(Distribution):
    """
    Any distribution from scipy.stats with overloading of addition and multiplication.

    Args:
            type (str): Data type of the distribution as a string. It must be one defined in `scipy.stats`
            loc (float): The location of the distribution (corresponds to mean in Gaussian)
            scale (float): The _scale of the distribution (corresponds to standard deviation in Gaussian)
            **kwargs:

    Examples:

        >>> d2 = ScipyDistribution("uniform")
        >>> print(d2.mean())
        0.5

        >>> try:
        ...     d = ScipyDistribution("something crazy")
        ... except ValueError as e:
        ...     print(str(e))
        Unknown distribution something crazy

    """

    def __init__(self, type: str, **kwargs) -> None:
        dist = getattr(stats, type.lower(), None)
        if dist is None:
            raise ValueError(f"Unknown distribution {type}")
        if "loc" not in kwargs.keys():
            kwargs["loc"] = 0.0
        if "scale" not in kwargs.keys():
            kwargs["scale"] = 1.0

        self._dist = dist(**kwargs)
        self._type = type

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result

    def type(self) -> str:
        return self._type

    def _make_dist(self, type: str, loc: float, scale: float):
        dist = getattr(stats, type.lower(), None)
        if dist is None:
            raise ValueError(f"Unknown distribution {type}")
        return dist(loc=loc, scale=scale)

    @property
    def loc(self):
        return self._dist.kwds.get("loc", 0.5)

    @property
    def scale(self):
        return self._dist.kwds.get("scale", 0.0)

    def mean(self) -> float:
        if self._type != "uniform":
            raise NotImplementedError(
                "Only uniform distributions are supported for now"
            )
        if self.scale < 1e-6:
            return float(self.loc)
        mymean = self._dist.mean()
        return float(mymean)

    def prob(self, val: float) -> float:
        """Returns the probability for the given value"""
        return self._dist.prob(val)

    def cum_prob(self, mn: float, mx: float) -> float:
        """Returns the probability for the given range"""
        return self._dist.cdf(mx) - self._dist.cdf(mn)

    def sample(self, size: int = 1) -> np.ndarray:
        return self._dist.rvs(size=size)

    @property
    def min(self):
        return self.loc - self.scale

    @property
    def max(self):
        return self.loc + self.scale

    def is_gaussian(self):
        return self._type == "normal"

    def is_uniform(self):
        return self._type == "uniform"

    def is_crisp(self) -> bool:
        """Returns true if this is a distribution with all probability at one point (delta(v))"""
        return self.scale < EPSILON / 1000

    def __call__(self, val: float) -> float:
        """Returns the probability for the given value"""
        return self.prob(val)

    def __add__(self, other):
        if isinstance(other, float) and self._type in ("uniform", "normal"):
            self._dist = self._make_dist(self._type, self.loc + other, self.scale)
        raise NotImplementedError()

    def __sub__(self, other):
        if isinstance(other, float) and self._type in ("uniform", "normal"):
            self._dist = self._make_dist(self._type, self.loc - other, self.scale)
        raise NotImplementedError()

    def __mul__(self, weight: float):
        if isinstance(weight, float) and self._type in ("uniform", "normal"):
            self._dist = self._make_dist(
                self._type, self.loc * weight, self.scale * weight
            )
        raise NotImplementedError()

    def __lt__(self, other) -> bool:
        """Check that a sample from `self` is ALWAYS less than a sample from other `other`"""
        if isinstance(other, numbers.Real):
            return self.max < float(other)
        return self.max < other.min

    def __le__(self, other) -> bool:
        """Check that a sample from `self` is ALWAYS less or equal a sample from other `other`"""
        return self < other or self == other

    # def __eq__(self, other):
    #     if isinstance(other, ScipyDistribution):
    #         return (
    #             self._type == other._type
    #             and abs(self.loc - other.loc) < EPSILON
    #             and abs(self.scale - other.scale) < EPSILON
    #         )
    #
    #     if isinstance(other, float):
    #         return abs(self.loc - other) < EPSILON and self.loc < EPSILON
    #     raise ValueError(f"Cannot compare Distribution with {type(other)}")

    def __eq__(self, other) -> bool:
        """Checks for equality of the two distributions"""
        if isinstance(other, numbers.Real):
            other = Real(other)
        return (
            self.loc == other.loc
            and self.scale == other.scale
            and (
                isinstance(other, Real)
                or isinstance(self, Real)
                or self.type == other.type
            )
        )

    def __ne__(self, other) -> bool:
        """Checks for ineqlaity of the distributions"""
        return not (self == other)

    def __gt__(self, other) -> bool:
        """Check that a sample from `self` is ALWAYS greater than a sample from other `other`"""
        if isinstance(other, numbers.Real):
            return self.min > other
        return self.min > other.max

    def __ge__(self, other) -> bool:
        """Check that a sample from `self` is ALWAYS greater or equal a sample from other `other`"""
        return self > other or self == other

    def __float__(self) -> float:
        """Converts to a float (usually by calling mean())"""
        return self.mean()

    def __str__(self):
        if self._type == "uniform":
            return f"U({self.loc}, {self.loc+self.scale})"
        if self._type == "normal":
            return f"G({self.loc}, {self.loc+self.scale})"
        return f"{self._type}(loc:{self.loc}, scale:{self.scale})"

    __repr__ = __str__


class UniformDistribution(ScipyDistribution):
    """A `ScipyDistribution` reprsenting a unifrom distribution"""

    def __init__(
        self, loc: float = 0.0, scale: float = 1.0, *, type: str = "uniform", **kwargs
    ) -> None:
        super().__init__(loc=loc, scale=scale, type="uniform", **kwargs)


class NormalDistribution(ScipyDistribution):
    """A `ScipyDistribution` reprsenting a unifrom distribution"""

    def __init__(
        self, loc: float = 0.0, scale: float = 1.0, *, type: str = "norm", **kwargs
    ) -> None:
        super().__init__(loc=loc, scale=scale, type="norm", **kwargs)


def uniform_around(
    value: float = 0.5,
    range: tuple[float, float] = (0.0, 1.0),
    uncertainty: float = 0.5,
    cls: type[Distribution] = ScipyDistribution,
) -> Distribution:
    """
    Generates a uniform distribution around the input value in the given range with given uncertainty

    Args:
        value: The value to generate the distribution around
        range: The range of possible values
        uncertainty: The uncertainty level required. 0.0 means no uncertainty and 1.0 means full uncertainty

    Returns:
        Distribution A uniform distribution around `value` with uncertainty (scale) `uncertainty`
    """
    if uncertainty >= 1.0:
        return cls(type="uniform", loc=range[0], scale=range[1])  # type: ignore
    if uncertainty <= 0.0:
        return Real(loc=value)
    scale = uncertainty * (range[1] - range[0])
    loc = max(range[0], (random.random() - 1.0) * scale + value)
    if loc + scale > range[1]:
        loc -= loc + scale - range[1]
    return cls(type="uniform", loc=loc, scale=scale)  # type: ignore



r"""A set of generic classes and corresponding functions to iterate and index them.


The main goal of having this module is to allow for flexible modeling of collections with *most* python builtin
collections (e.g. `list`, `tuple`, `generator`) or even `Callable`\ s without awkward `isinstance` calls in
other modules of the library.

"""
from __future__ import annotations
from typing import Any, Callable, Iterable, Mapping, Sequence, Union

__all__ = [
    "IterableMapping",  # A mapping combining dicts, lists, tuples, and generators
    "GenericMapping",
    # An iterable mapping or a callable (e.g. function/functor)
    # GenericMapping functions (apply to GMapping and IterableMapping)
    "gmap",  # An index on an IterableMapping or a call on a Callable
    "gget",
    # Like `gmap` but can return a default value
    # IterableMapping functions (apply only to IterableMappings)
    "ienumerate",  # enumerates a mapping as key-value pairs
    "iitems",  # enumerates a mapping as key-value pairs
    "ivalues",  # enumerates all values in a mapping (no keys)
    "ikeys",  # enumerates all keys in a mapping (no values)
    "iget",  # Similar to `gget` but applies only to IterableMappings
]
# Generic classes that work with a variety of python mappings like collections or callables
IterableMapping = Union[Mapping, Sequence]
"""Something that can be iterated upon with Key-value pairs (e.g. list, dict, tuple)."""
GenericMapping = Union[Callable[[Any], Any], IterableMapping]
"""Something that can be indexed using [] or called using ()"""


def gmap(group: GenericMapping, param: Any) -> Any:
    """Calls or indexes the group by the param

    Args:
        group: Either a Callable or a Mapping
        param: The parameters to use for mapping

    Examples:

        >>> gmap([1, 23, 44], 1)
        23
        >>> gmap({"a": 3, "b": 5, "c": 4}, "c")
        4
        >>> gmap(lambda x: 3 * x, 20)
        60

    Returns:

    """

    if hasattr(group, "__call__"):
        return group(param)  # type: ignore

    return group[param]  # type: ignore


def iget(x: IterableMapping, _key: Any, default=None) -> Any:
    """Get an item from an IterableMapping

    Args:
        x: the generic mapping
        _key: key (must be immutable)
        default: default value if no value attached with the key is found

    Examples:

        Example with a list

        >>> [iget([10, 20, 30], _) for _ in (0, 2, 1, -1, 4)]
        [10, 30, 20, 30, None]

        Example with a dictionary

        >>> [iget({"a": 10, "b": 20, "c": 30}, _) for _ in ("a", "c", "b", -1, "d")]
        [10, 30, 20, None, None]

        Example with a tuple

        >>> [iget((10, 20, 30), _) for _ in (0, 2, 1, -1, 4)]
        [10, 30, 20, 30, None]

        Example with a generator

        >>> [iget(range(10, 40, 10), _) for _ in (0, 2, 1, -1, 4)]
        [10, 30, 20, 30, None]

    Returns:

    """

    if isinstance(x, dict):
        return x.get(_key, default)

    try:
        return x[_key]  # type: ignore
    except IndexError:
        return default


def gget(x: GenericMapping, _key: Any, default=None) -> Any:
    """Get an item from an IterableMapping

    Args:
        x: the generic mapping
        _key: key (must be immutable)
        default: default value if no value attached with the key is found

    Examples:

        Example with a list

        >>> [gget([10, 20, 30], _) for _ in (0, 2, 1, -1, 4)]
        [10, 30, 20, 30, None]

        Example with a dictionary

        >>> [gget({"a": 10, "b": 20, "c": 30}, _) for _ in ("a", "c", "b", -1, "d")]
        [10, 30, 20, None, None]

        Example with a tuple

        >>> [gget((10, 20, 30), _) for _ in (0, 2, 1, -1, 4)]
        [10, 30, 20, 30, None]

        Example with a generator

        >>> [gget(range(10, 40, 10), _) for _ in (0, 2, 1, -1, 4)]
        [10, 30, 20, 30, None]

    Returns:

    """

    if hasattr(x, "apply"):
        # noinspection PyBroadException
        try:
            return x(_key)  # type: ignore

        except Exception:
            return default

    else:
        return iget(x, _key, default)  # type: ignore


def ienumerate(x: IterableMapping) -> Iterable[tuple[Any, Any]]:
    """Enumerates a GenericMapping.

    Args:
        x (IterableMapping): A generic mapping (see `GenericMapping`)

    Examples:

        Example with a list

        >>> for k, cutoff_utility in ienumerate([10, 20, 30]):
        ...     print(k, cutoff_utility, end="-")
        0 10-1 20-2 30-

        Example with a dictionary

        >>> for k, cutoff_utility in ienumerate({"a": 10, "b": 20, "c": 30}):
        ...     print(k, cutoff_utility, end="-")
        a 10-b 20-c 30-

        Example with a tuple

        >>> for k, cutoff_utility in ienumerate((10, 20, 30)):
        ...     print(k, cutoff_utility, end="-")
        0 10-1 20-2 30-

        Example with a generator

        >>> for k, cutoff_utility in ienumerate(range(10, 40, 10)):
        ...     print(k, cutoff_utility, end="-")
        0 10-1 20-2 30-

    Returns:
        a generator/iterator with tuples of key-value pairs.

    """

    if isinstance(x, dict):
        return x.items()

    else:
        return enumerate(x)


iitems = ienumerate


def ivalues(x: IterableMapping) -> Iterable[Any]:
    """Returns all keys of the iterable.

    Args:
        x (IterableMapping): A generic mapping (see `GenericMapping`)

    Examples:

        Example with a list

        >>> for k in ivalues([10, 20, 30]):
        ...     print(k, end="-")
        10-20-30-

        Example with a dictionary

        >>> for k in ivalues({"a": 10, "b": 20, "c": 30}):
        ...     print(k, end="-")
        10-20-30-

        Example with a tuple

        >>> for k in ivalues((10, 20, 30)):
        ...     print(k, end="-")
        10-20-30-

        Example with a generator

        >>> for k in ivalues(range(10, 40, 10)):
        ...     print(k, end="-")
        10-20-30-

    Returns:
        a generator/iterator with tuples of key-value pairs.

    """

    if isinstance(x, dict):
        return list(x.values())

    else:
        return x


def ikeys(x: IterableMapping) -> Iterable[Any]:
    """Returns all keys of the iterable.

    Args:
        x (IterableMapping): A generic mapping (see `GenericMapping`)

    Examples:

        Example with a list

        >>> for k in ikeys([10, 20, 30]):
        ...     print(k, end="-")
        0-1-2-

        Example with a dictionary

        >>> for k in ikeys({"a": 10, "b": 20, "c": 30}):
        ...     print(k, end="-")
        a-b-c-

        Example with a tuple

        >>> for k in ikeys((10, 20, 30)):
        ...     print(k, end="-")
        0-1-2-

        Example with a generator

        >>> for k in ikeys(range(10, 40, 10)):
        ...     print(k, end="-")
        0-1-2-

    Returns:
        a generator/iterator with tuples of key-value pairs.

    """

    if isinstance(x, dict):
        return list(x.keys())
    else:
        return range(len(x))  # type: ignore



"""Defines basic config for NEGMAS"""
from __future__ import annotations

import json
from os import environ
from pathlib import Path

__all__ = [
    "NEGMAS_CONFIG",
    "CONFIG_KEY_JNEGMAS_JAR",
    "CONFIG_KEY_GENIUS_BRIDGE_JAR",
    "negmas_config",
]

LOCAL_NEGMAS_CONFIG_FILENAME = "negmasconf.json"

NEGMAS_DEFAULT_PATH = Path(
    environ.get("NEGMAS_DEFAULT_PATH", Path.home() / "negmas" / "config.json")
)
"""Default path for NegMAS configurations"""

CONFIG_KEY_JNEGMAS_JAR = "jnegmas_jar"
"""Key name for the JNegMAS jar in `NEGMAS_CONFIG`"""

CONFIG_KEY_GENIUS_BRIDGE_JAR = "genius_bridge_jar"
"""Key name for the Genius bridge jar in `NEGMAS_CONFIG`"""

CONFIG_KEY_WARN_SLOW_OPS = "warn_slow_ops"

NEGMAS_CONFIG = {
    CONFIG_KEY_JNEGMAS_JAR: str(NEGMAS_DEFAULT_PATH.parent / "files" / "jnegmas.jar"),
    CONFIG_KEY_GENIUS_BRIDGE_JAR: str(
        NEGMAS_DEFAULT_PATH.parent / "files" / "geniusbridge.jar"
    ),
    CONFIG_KEY_WARN_SLOW_OPS: 100_000_000,
}

# loading config file if any
__conf_path = Path(NEGMAS_DEFAULT_PATH).expanduser().absolute()

if __conf_path.exists():
    try:
        with open(__conf_path) as f:
            NEGMAS_CONFIG.update(json.load(f))
    except Exception:
        pass

local_path = Path.cwd() / LOCAL_NEGMAS_CONFIG_FILENAME
if local_path.exists():
    try:
        with open(local_path) as f:
            NEGMAS_CONFIG.update(json.load(f))
    except Exception:
        pass


def _from_env(key: str, default):
    envkey = "NEGMAS_" + key.upper()
    v = environ.get(envkey, default)
    if key in (CONFIG_KEY_WARN_SLOW_OPS,):
        return int(v) if v else 0
    return v


def negmas_config(key: str, default):
    """
    Returns the config value associated with the given key.


    Remarks:
        - config values are read from the following sources (in descending order of priority):
            - Environment variable with the name NEGMAS_{key} (with the key converted to all uppercase)
            - Local file called negmasconf.json (with the key all lowercase)
            - json file stored at the location indicated by environment variable "NEGMAS_DEFAULT_PATH" (with the key all lowercase)
            - ~/negmas/config.json (with the key all lowercase)
            - A default value hardcoded in the negmas library. For paths, this usually lies under ~/negmas
    """
    return _from_env(key, NEGMAS_CONFIG.get(key.lower(), default))



#!/usr/bin/env python
"""
Extra helpers
"""

from __future__ import annotations

import itertools
import random
import socket
from math import exp, log
from typing import Any, Callable, Iterable, Sequence

import numpy as np

from ..protocols import HasMinMax

__all__ = [
    "monotonic_minmax",
    "monotonic_multi_minmax",
    "nonmonotonic_multi_minmax",
    "nonmonotonic_minmax",
    "make_callable",
    "get_free_tcp_port",
    "intin",
    "floatin",
    "distribute_integer_randomly",
    "generate_random_weights",
]


def generate_random_weights(n):
    """
    Generates a list of n random weights between 0 and 1 that sum to 1.

    Args:
      n: The number of weights to generate.

    Returns:
      A list of n random weights between 0 and 1 that sum to 1.
    """

    # Generate n random numbers between 0 and 1 using the Dirichlet distribution,
    # which ensures equal probability of generating any valid combination of weights.
    weights = np.random.dirichlet(np.ones(n))

    return weights


def distribute_integer_randomly(
    n: int, m: int, min_per_bin: int | None = 1
) -> list[int]:
    """
    Distributes an integer n over a list of m values randomly, with each value at least one.

    Args:
      n: The integer to distribute.
      m: The number of values to distribute over.
      min_per_bin: Minimum number of elements per bin. This is only guaranteed if n >= m * min_per_bin.
                   If None, then the distribution will be as normal as possible

    Returns:
      A list of m integers, where each value is at least one.

    Remarks:
        - if n < m * min_per_bin, n will be distributed randomly with some zeros inserted.

    """
    if m < 1 and n != 0 and min_per_bin is not None:
        raise ValueError(f"Cannot distribute {n} over {m} bins")
    if n == 0:
        return [0] * m

    # if n < m:
    #     raise ValueError(f"Cannot distribute {n} over {m} bins.")
    # just distribute as much as possible if min_per_bin cannot be achieved
    if min_per_bin is not None and n < m * min_per_bin:
        n_active = int(n // min_per_bin)
        assert (
            0 <= n_active <= n
        ), f"Distributing {n} over {m} with {min_per_bin=} leads to {n_active} values which should be between {0} and {n}!!"
        lst = [min_per_bin] * n_active + [0] * (m - n_active)
        random.shuffle(lst)
        return lst
    # Find the min_per_bin if equal distribution is asked for
    equal_dist = False
    if min_per_bin is None:
        min_per_bin = int(n // m)
        equal_dist = True
    # Calculate base distribution and remainder
    base_distribution = int(min_per_bin)
    remainder = n - m * base_distribution

    # Create a list with base distribution for each value
    distribution = [base_distribution] * m

    # Add remainder to the first `remainder` elements randomly
    for i in range(remainder):
        nxt = random.randrange(m) if not equal_dist else (i % m)
        distribution[nxt] += 1

    # # Shuffle the list to randomize remainder distribution
    # random.shuffle(distribution)

    return distribution


def floatin(
    x: float | tuple[float, float] | Sequence[float], log_uniform: bool
) -> float:
    """Samples a value. A random choice is made if x is another sequence, a value between the two limits is used if the input is a tuple otherwise x is returned"""
    if isinstance(x, tuple) and len(x) == 2:
        x = tuple(float(_) for _ in x)
        if x[0] == x[-1]:
            return x[0]
        if log_uniform:
            L = [log(_) for _ in x]
            return min(x[1], max(x[0], exp(random.random() * (L[1] - L[0]) + L[0])))

        return random.random() * (x[1] - x[0]) + x[0]
    if isinstance(x, Sequence):
        return float(random.choice(x))
    return float(x)


def intin(x: int | tuple[int, int] | Sequence[int], log_uniform: bool = False) -> int:
    """Samples a value. A random choice is made if x is another sequence, a value between the two limits is used if the input is a tuple otherwise x is returned"""
    if isinstance(x, tuple) and len(x) == 2:
        x = tuple(int(_) for _ in x)
        if x[0] == x[-1]:
            return int(x[0])
        if log_uniform:
            L = [log(_) for _ in x]
            return min(
                x[1], max(x[0], int(0.5 + exp(random.random() * (L[1] - L[0]) + L[0])))
            )

        if x[0] == x[1]:
            return int(x[0])
        return random.randint(*x)
    if isinstance(x, Sequence):
        return int(random.choice(x))
    return int(x)


def nonmonotonic_minmax(
    input: Iterable, f: Callable[[Any], float]
) -> tuple[float, float]:
    """Finds the limits of a function `f` for the input assuming that it is non-monotonic and input is iterable"""
    mn, mx = float("inf"), float("-inf")
    fmn, fmx = float("-inf"), float("inf")
    for x in input:
        fx = float(f(x))
        if fx < mn:
            mn, fmn = x, fx
        if fx > mx:
            mx, fmx = x, fx
    return fmn, fmx


def nonmonotonic_multi_minmax(
    input: Iterable[Iterable], f: Callable[[Any], float]
) -> tuple[float, float]:
    """Finds the limits of a function `f` for each input assuming that it is non-monotonic and input is iterable"""
    return nonmonotonic_minmax(itertools.product(*input), f)


def monotonic_minmax(
    input: HasMinMax, f: Callable[[Any], float]
) -> tuple[float, float]:
    """Finds the limits of a function `f` for the input assuming that it is monotonic and input has `min_value` and `max_value` members"""
    a, b = input.min_value, input.max_value
    fa, fb = f(a), f(b)
    if fb < fa:
        return float(fb), float(fa)
    return float(fa), float(fb)


def monotonic_multi_minmax(
    input: Iterable[HasMinMax], f: Callable[[Any], float]
) -> tuple[float, float]:
    """Finds the limits of a function `f` for the input assuming that it is monotonic and each input has `min_value` and `max_value` members"""
    vals = [(i.min_value, i.max_value) for i in input]
    return nonmonotonic_minmax(itertools.product(*vals), f)


def remove_qoutes(s: str) -> str:
    return s.replace('"', "~`")


def recover_qoutes(s: str) -> str:
    return s.replace("~`", '"')


def make_callable(x: dict | Sequence | Callable | None) -> Callable:
    """
    Converts its input to a callable (i.e. can be called using () operator).


    Examples:

        >>> make_callable(lambda x: 2 * x)(4)
        8

        >>> make_callable(dict(a=1, b=3))("a")
        1

        >>> make_callable((3, 4, 5))(1)
        4
    """
    if x is None:
        return lambda a: a
    if isinstance(x, Callable):
        return x
    return lambda a: x[a]


def get_free_tcp_port():
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(("", 0))
    _, port = tcp.getsockname()
    tcp.close()
    return port



from __future__ import annotations
from typing import Any, Iterable

import numpy as np

from ..common import Value

np.seterr(all="raise")  # setting numpy to raise exceptions in case of errors


__all__ = ["_loc", "_locs", "_scale", "_upper", "_uppers", "argmax", "argmin", "argmin"]


def _loc(u: Value):
    """Returns the lower bound of a Value"""
    return u if isinstance(u, float) else u.loc


def _locs(us: Iterable[Value]):
    """Returns the lower bound of an iterable of Value(s)"""
    return [u if isinstance(u, float) else u.loc for u in us]


def _scale(u: Value):
    """Returns the difference between the upper and lower bounds of a Value"""
    return 0.0 if isinstance(u, float) else u.scale


def _upper(u: Value):
    """Returns the upper bound of a Value"""
    return u if isinstance(u, float) else (u.loc + u.scale)


def _uppers(us: Iterable[Value]):
    """Returns the upper bounds of an Iterble of Values"""
    return [u if isinstance(u, float) else (u.loc + u.scale) for u in us]


def argmax(iterable: Iterable[Any]):
    """Returns the index of the maximum"""
    return max(enumerate(iterable), key=lambda x: x[1])[0]


def argsort(iterable: Iterable[Any]):
    """Returns a list of indices that would sort the iterable"""
    return [_[0] for _ in sorted(enumerate(iterable), key=lambda x: x[1])]


def argmin(iterable):
    """Returns the index of the minimum"""
    return min(enumerate(iterable), key=lambda x: x[1])[0]



from __future__ import annotations
from negmas.outcomes.ordinal_issue import DiscreteOrdinalIssue, OrdinalIssue

__all__ = ["CardinalIssue"]


class CardinalIssue(OrdinalIssue):
    """
    An `Issue` for which differences between values are meaningful.
    """

    def __init__(self, values, name=None) -> None:
        super().__init__(values, name)
        try:
            for _ in range(3):
                self.rand_valid() - self.rand_valid()  # type: ignore
        except Exception as e:
            raise ValueError(
                f"Cardinal issues should support subtraction between issue values. {self.value_type} does not ({e})"
            )


class DiscreteCardinalIssue(DiscreteOrdinalIssue, CardinalIssue):
    """
    An issue that has an ordering and for which differences between values is defined (i.e. subtraction)
    """



from __future__ import annotations
from negmas.common import MechanismState

from ..common import GBResponse, ResponseType, ThreadState
from ..evaluators.base import LocalEvaluationStrategy

__all__ = ["GAOEvaluationStrategy"]


class GAOEvaluationStrategy(LocalEvaluationStrategy):
    def eval(
        self,
        negotiator_id: str,
        state: ThreadState,
        history: list[ThreadState],
        mechanism_state: MechanismState,
    ) -> GBResponse:
        offer, responses = state.new_offer, state.new_responses
        if offer is None:
            return None
        if any(_ == ResponseType.END_NEGOTIATION for _ in responses.values()):
            return None
        if all(_ == ResponseType.ACCEPT_OFFER for _ in responses.values()):
            return offer
        return "continue"



from .const import *
from .linear import *
from .mapping import *
from .nonlinear import *
from .random_ufun import *
from .rankonly_ufun import *


__all__ = (
    const.__all__
    + linear.__all__
    + mapping.__all__
    + nonlinear.__all__
    + random_ufun.__all__
    + rankonly_ufun.__all__
)



from __future__ import annotations
from pytest import mark

from negmas import MappingUtilityFunction, SorterNegotiator
from negmas.ga import GAMechanism


@mark.parametrize("n_negotiators,n_outcomes", [(2, 10), (3, 50), (2, 50), (3, 5)])
def test_ga_mechanism(n_negotiators, n_outcomes):
    mechanism = GAMechanism(outcomes=n_outcomes, n_steps=3)
    ufuns = MappingUtilityFunction.generate_random(n_negotiators, outcomes=n_outcomes)
    for i in range(n_negotiators):
        mechanism.add(SorterNegotiator(name=f"agent{i}"), preferences=ufuns[i])
    assert mechanism.state.step == 0
    mechanism.step()
    assert mechanism.state.step == 1
    assert mechanism.dominant_outcomes is not None
    mechanism.run()



"""
Implements Stacked Alternating Offers (SAO) mechanism.
"""

from __future__ import annotations

import functools
import sys
import time
from collections import defaultdict
from typing import TYPE_CHECKING

from attr import asdict
from rich import print

from negmas import warnings
from negmas.gb.negotiators import GBNegotiator
from negmas.helpers.strings import humanize_time

from ..common import TraceElement
from ..events import Event
from ..helpers import TimeoutCaller, TimeoutError, exception2str
from ..mechanisms import Mechanism, MechanismStepResult
from ..outcomes.common import Outcome
from ..outcomes.outcome_ops import cast_value_types, outcome_types_are_ok
from .common import SAONMI, ResponseType, SAOResponse, SAOState
from .negotiators import SAONegotiator

if TYPE_CHECKING:
    from negmas.preferences import Preferences


__all__ = ["SAOMechanism", "SAOProtocol", "TraceElement"]

DEFAULT_COLORMAP = "jet"


class SAOMechanism(
    Mechanism[SAONMI, SAOState, SAOResponse, SAONegotiator | GBNegotiator]
):
    """
    Implements Several variants of the Stacked Alternating Offers Protocol

    Args:
        outcome_space: The negotiation agenda
        issues: A list of issues defining the outcome-space of the negotiation
        outcomes: A list of outcomes defining the outcome-space of the negotiation
        n_steps: The maximum number of negotiaion rounds (see `time_limit` )
        time_limit: The maximum wall-time allowed for the negotiation (see `n_steps`)
        step_time_limit: The maximum wall-time allowed for a single negotiation round.
        max_n_agents: The maximum number of negotiators allowed to join the negotiation.
        dynamic_entry: Whether it is allowed for negotiators to join the negotiation after it starts
        cache_outcomes: If true, the mechnism will catch `outcomes` and a discrete version (`discrete_outcomes`) that
                        can be accessed by any negotiator through their AMI.
        max_cardinality: Maximum number or outcomes to use when disctetizing the outcome-space
        annotation: A key-value mapping to keep around. Accessible through the AMI but not used by the mechanism.
        end_on_no_response: End the negotiation if any negotiator returns NO_RESPONSE from `respond`/`counter` or returns
                            REJECT_OFFER then refuses to give an offer (by returning `None` from `proposee/`counter`).
        enable_callbacks: Enable callbacks like on_round_start, etc. Note that on_negotiation_end is always received
                          by the negotiators no matter what is the setting for this parameter.
        check_offers: If true, offers are checked to see if they are valid for the negotiation
                      outcome-space and if not the offer is considered None which is the same as
                      refusing to offer (NO_RESPONSE).
        enforce_issue_types: If True, the type of each issue is enforced depending on the value of `cast_offers`
        cast_offers: If true, each issue value is cast using the issue's type otherwise an incorrect type will be considered an invalid offer. See `check_offers`. Only
                     used if `enforce_issue_types`
        ignore_negotiator_exceptions: just silently ignore negotiator exceptions and consider them no-responses.
        offering_is_accepting: Offering an outcome implies accepting it. If not, the agent who proposed an offer will
                               be asked to respond to it after all other agents.
        sync_calls: If given, calls to negotiators will be synchronized. This will not enforce timeouts on
                    single calls which means that a negotiator in an infinite loop will hog the CPU. By default
                    calls are done using a different thread that is killed when the timeout passes. This may, but
                    is not guaranteed to, resolve this issue at the expense of slower negotiations and harder debugging
        name: Name of the mecnanism
        **kwargs: Extra paramters passed directly to the `Mechanism` constructor

    Remarks:

        - One and only one of `outcome_space`, `issues`, `outcomes` can be given
        - If both `n_steps` and `time_limit` are passed, the negotiation ends when either of the limits is reached.
        - Negotiations may take longer than `time_limit` because negotiators are not interrupted while they are
          executing their `respond` or `propose` methods.

    Events:

        - negotiator_exception: Data=(negotiator, exception) raised whenever a negotiator raises an exception if
          ignore_negotiator_exceptions is set to True.
    """

    def __init__(
        self,
        dynamic_entry=False,
        extra_callbacks=True,
        end_on_no_response=True,
        avoid_ultimatum=False,
        check_offers=False,
        enforce_issue_types=False,
        cast_offers=False,
        ignore_negotiator_exceptions=False,
        offering_is_accepting=True,
        allow_offering_just_rejected_outcome=True,
        name: str | None = None,
        max_wait: int = sys.maxsize,
        sync_calls: bool = False,
        initial_state: SAOState | None = None,
        one_offer_per_step: bool = False,
        **kwargs,
    ):
        debug = kwargs.get("debug", False)
        if debug:
            sync_calls = True
        if avoid_ultimatum:
            warnings.warn(
                "Support for Avoid-Ultimatum will be removed soon. We will force avoid_ultimatum to False",
                warnings.NegmasWarning,
            )
            avoid_ultimatum = False
        super().__init__(
            dynamic_entry=dynamic_entry,
            extra_callbacks=extra_callbacks,
            initial_state=SAOState() if not initial_state else initial_state,
            name=name,
            **kwargs,
        )
        self.nmi = SAONMI(
            **{
                **asdict(self.nmi, recurse=False),
                **dict(
                    end_on_no_response=end_on_no_response,
                    one_offer_per_step=one_offer_per_step,
                ),
            }
        )
        assert self.nmi.end_on_no_response == end_on_no_response
        assert self.nmi.one_offer_per_step == one_offer_per_step
        self._one_offer_per_step = one_offer_per_step
        self._current_state: SAOState

        # self._history: list[SAOState] = []
        n_steps, time_limit = self.n_steps, self.time_limit
        if (n_steps is None or n_steps == float("inf")) and (
            time_limit is None or time_limit == float("inf")
        ):
            warnings.warn(
                "You are passing no time_limit and no n_steps to an SAOMechanism. The mechanism may never finish!!",
                warnings.NegmasInfiniteNegotiationWarning,
            )
        self._sync_calls = sync_calls
        self.params["one_offer_per_step"] = one_offer_per_step
        self.params["end_on_no_response"] = end_on_no_response
        self.params["enable_callbacks"] = extra_callbacks
        self.params["sync_calls"] = sync_calls
        self.params["check_offers"] = check_offers
        self.params["offering_is_accepting"] = offering_is_accepting
        self.params["enforce_issue_types"] = enforce_issue_types
        self.params["cast_offers"] = cast_offers
        self.params[
            "allow_offering_just_rejected_outcome"
        ] = allow_offering_just_rejected_outcome
        self._n_max_waits = max_wait if max_wait is not None else float("inf")
        self.params["max_wait"] = self._n_max_waits
        self.ignore_negotiator_exceptions = ignore_negotiator_exceptions
        self.allow_offering_just_rejected_outcome = allow_offering_just_rejected_outcome
        self.end_negotiation_on_refusal_to_propose = end_on_no_response
        self.check_offers = check_offers
        self._enforce_issue_types = enforce_issue_types
        self._cast_offers = cast_offers

        self._last_checked_negotiator = -1
        self._current_proposer = None
        self._frozen_neg_list = None
        self._no_responses = 0
        self._offering_is_accepting = offering_is_accepting
        self._n_waits = 0
        self._waiting_time: dict[str, float] = defaultdict(float)
        self._waiting_start: dict[str, float] = defaultdict(lambda: float("inf"))
        self._selected_first = 0

    @property
    def state(self) -> SAOState:
        """Returns the current state.

        Override `extra_state` if you want to keep extra state
        """
        return self._current_state

    def add(  #
        self,
        negotiator: SAONegotiator | GBNegotiator,
        *,
        preferences: Preferences | None = None,
        role: str | None = None,
        **kwargs,
    ) -> bool | None:
        from ..genius.negotiator import GeniusNegotiator

        added = super().add(negotiator, preferences=preferences, role=role, **kwargs)
        if (
            added
            and isinstance(negotiator, GeniusNegotiator)
            and self.nmi.time_limit is not None
            and self.nmi.time_limit != float("inf")
            and self.nmi.n_steps is not None
            and self.nmi.n_steps != float("inf")
        ):
            warnings.warn(
                f"{negotiator.id} of type {negotiator.__class__.__name__} is joining "
                f"SAOMechanism which has a time_limit of {self.nmi.time_limit} seconds "
                f"and a n_steps of {self.nmi.n_steps}. This agnet will only know about the "
                f"time_limit and will not know about the n_steps!!!",
                warnings.NegmasStepAndTimeLimitWarning,
            )
        return added

    def set_sync_call(self, v: bool):
        self._sync_call = v

    def _agent_info(self):
        state = self._current_state
        current_proposer_agent = (
            self._current_proposer.owner if self._current_proposer else None
        )
        if current_proposer_agent:
            current_proposer_agent = current_proposer_agent.id
        new_offerer_agents = []
        for neg_id, _ in state.new_offers:
            neg = self._negotiator_map.get(neg_id, None)
            agent = neg.owner if neg else None
            if agent is not None:
                new_offerer_agents.append(agent.id)
            else:
                new_offerer_agents.append(None)
        return current_proposer_agent, new_offerer_agents

    def next_negotitor_ids(self) -> list[str]:
        """Returns a list of negotiator IDs in the order they are to be run in the next call to step()"""
        n_negotiators = len(self.negotiators)
        if self._frozen_neg_list is not None:
            ordered_indices = self._frozen_neg_list
        else:
            ordered_indices = [
                (_ + self._last_checked_negotiator + 1) % n_negotiators
                for _ in range(n_negotiators)
            ]

        if ordered_indices and self._one_offer_per_step:
            ordered_indices = ordered_indices[:1]
        ids = self.negotiator_ids
        return [ids[_] for _ in ordered_indices]

    def _stop_waiting(self, negotiator_id):
        self._waiting_time[negotiator_id] = 0.0
        self._waiting_start[negotiator_id] = float("inf")
        self._n_waits = 0
        self._frozen_neg_list = None

    def __call__(self, state, action=None) -> MechanismStepResult:
        """
        implements a round or a single step of the Stacked Alternating Offers Protocol.

        Args:
            state: Current state of the mechanism
            action: The action to use as a mapping from negotiator ID (key) to its response (value).
                    If not given, the negotiator(s) is called to generate its response.
        """
        state = self._current_state
        if self._frozen_neg_list is None:
            state.new_offers = []
        negotiators: list[SAONegotiator | GBNegotiator] = self.negotiators
        n_negotiators = len(negotiators)
        # times = dict(zip([_.id for _ in negotiators], itertools.repeat(0.0)))
        times = defaultdict(float, self._waiting_time)
        exceptions = dict(
            zip([_.id for _ in negotiators], [list() for _ in negotiators])
        )

        def _safe_counter(
            negotiator, *args, **kwargs
        ) -> tuple[SAOResponse | None, bool]:
            assert (
                not state.waiting or negotiator.id == state.current_proposer
            ), f"We are waiting with {state.current_proposer} as the last offerer but we are asking {negotiator.id} to offer\n{state}"
            if self.verbosity > 2:
                print(
                    f"{self.name}: {negotiator.name} called after {humanize_time(time.perf_counter() - self._start_time, show_ms=True) if self._start_time else 0}",
                    flush=True,
                )
            rem = self.remaining_time
            if rem is None:
                rem = float("inf")
            timeout = min(
                self.nmi.negotiator_time_limit - times[negotiator.id],
                self.nmi.step_time_limit,
                rem,
                self._hidden_time_limit - self.time,
            )
            given_response = action.pop(negotiator.id, None) if action else None
            if timeout is None or timeout == float("inf") or self._sync_calls:
                __strt = time.perf_counter()
                try:
                    if (
                        negotiator == self._current_proposer
                    ) and self._offering_is_accepting:
                        self._current_state.n_acceptances = 0
                        response = (
                            given_response
                            if given_response
                            else negotiator(*args, **kwargs)
                        )
                    else:
                        response = (
                            given_response
                            if given_response
                            else negotiator(*args, **kwargs)
                        )
                except TimeoutError:
                    response = None
                    try:
                        negotiator.cancel()
                    except Exception:
                        pass
                except Exception as ex:
                    exceptions[negotiator.id].append(exception2str())
                    if self.ignore_negotiator_exceptions:
                        self.announce(
                            Event(
                                "negotiator_exception",
                                {"negotiator": negotiator, "exception": ex},
                            )
                        )
                        times[negotiator.id] += time.perf_counter() - __strt
                        return SAOResponse(ResponseType.END_NEGOTIATION, None), True
                    else:
                        raise ex
                times[negotiator.id] += time.perf_counter() - __strt
            else:
                fun = functools.partial(negotiator, *args, **kwargs)
                __strt = time.perf_counter()
                try:
                    if (
                        negotiator == self._current_proposer
                    ) and self._offering_is_accepting:
                        state.n_acceptances = 0
                        response = (
                            given_response
                            if given_response
                            else TimeoutCaller.run(fun, timeout=timeout)
                        )
                    else:
                        response = (
                            given_response
                            if given_response
                            else TimeoutCaller.run(fun, timeout=timeout)
                        )
                except TimeoutError:
                    response = None
                except Exception as ex:
                    exceptions[negotiator.id].append(exception2str())
                    if self.ignore_negotiator_exceptions:
                        self.announce(
                            Event(
                                "negotiator_exception",
                                {"negotiator": negotiator, "exception": ex},
                            )
                        )
                        times[negotiator.id] += time.perf_counter() - __strt
                        return SAOResponse(ResponseType.END_NEGOTIATION, None), True
                    else:
                        raise ex
                times[negotiator.id] += time.perf_counter() - __strt
            if (
                self.check_offers
                and response is not None
                and response.outcome is not None
            ):
                if not self.outcome_space.is_valid(response.outcome):
                    return SAOResponse(response.response, None), False
                # todo: do not use .issues here as they are not guaranteed to exist (if it is not a cartesial outcome space)
                if self._enforce_issue_types and hasattr(self.outcome_space, "issues"):
                    if outcome_types_are_ok(
                        response.outcome, getattr(self.outcome_space, "issues")
                    ):
                        return response, False
                    elif self._cast_offers:
                        return (
                            SAOResponse(
                                response.response,
                                cast_value_types(
                                    response.outcome,
                                    getattr(self.outcome_space, "issues"),
                                ),
                            ),
                            False,
                        )
                    return SAOResponse(response.response, None), False
            return response, False

        proposers, proposer_indices = [], []
        for i, neg in enumerate(negotiators):
            if not neg.capabilities.get("propose", False):
                continue
            proposers.append(neg)
            proposer_indices.append(i)
        n_proposers = len(proposers)
        if n_proposers < 1:
            if not self.dynamic_entry:
                state.broken = True
                state.has_error = True
                state.error_details = "No proposers and no dynamic entry"
                return MechanismStepResult(state, times=times, exceptions=exceptions)
            else:
                return MechanismStepResult(state, times=times, exceptions=exceptions)
        if self._frozen_neg_list is not None:
            ordered_indices = self._frozen_neg_list
        else:
            ordered_indices = [
                (_ + self._last_checked_negotiator + 1) % n_negotiators
                for _ in range(n_negotiators)
            ]

        if ordered_indices and self._one_offer_per_step:
            ordered_indices = ordered_indices[:1]

        for _, neg_indx in enumerate(ordered_indices):
            self._last_checked_negotiator = neg_indx
            neg = self.negotiators[neg_indx]
            strt = time.perf_counter()
            resp, has_exceptions = _safe_counter(neg, state=self.state)
            self._negotiator_times[neg.id] += time.perf_counter() - strt
            if has_exceptions:
                state.broken = True
                state.has_error = True
                state.error_details = str(exceptions[neg.id])
                return MechanismStepResult(state, times=times, exceptions=exceptions)
            if resp is None:
                state.timedout = True
                return MechanismStepResult(state, times=times, exceptions=exceptions)
            if resp.response == ResponseType.WAIT:
                self._waiting_start[neg.id] = min(self._waiting_start[neg.id], strt)
                self._waiting_time[neg.id] += time.perf_counter() - strt
                self._last_checked_negotiator = (neg_indx - 1) % n_negotiators
                offered = {self._negotiator_index[_[0]] for _ in state.new_offers}
                did_not_offer = sorted(
                    list(set(range(n_negotiators)).difference(offered))
                )
                assert neg_indx in did_not_offer
                indx = did_not_offer.index(neg_indx)
                assert (
                    self._frozen_neg_list is None
                    or self._frozen_neg_list[0] == neg_indx
                )
                self._frozen_neg_list = did_not_offer[indx:] + did_not_offer[:indx]
                self._n_waits += 1
            else:
                self._stop_waiting(neg.id)

            if resp is None or time.perf_counter() - strt > self.nmi.step_time_limit:
                state.timedout = True
                return MechanismStepResult(state, times=times, exceptions=exceptions)
            if self._extra_callbacks:
                if state.current_offer is not None:
                    for other in self.negotiators:
                        if other is not neg:
                            other.on_partner_response(
                                state=self.state,
                                partner_id=neg.id,
                                outcome=state.current_offer,
                                response=resp.response,
                            )
            if resp.response == ResponseType.NO_RESPONSE:
                continue
            if resp.response == ResponseType.WAIT:
                if self._n_waits > self._n_max_waits:
                    self._stop_waiting(neg.id)
                    state.timedout = True
                    state.waiting = False
                    return MechanismStepResult(
                        state, times=times, exceptions=exceptions
                    )
                state.waiting = True
                return MechanismStepResult(state, times=times, exceptions=exceptions)
            if resp.response == ResponseType.END_NEGOTIATION:
                state.broken = True
                return MechanismStepResult(state, times=times, exceptions=exceptions)
            if resp.response == ResponseType.ACCEPT_OFFER:
                state.n_acceptances += 1
                if state.n_acceptances == n_negotiators:
                    state.agreement = self._current_state.current_offer
                    return MechanismStepResult(
                        state,
                        timedout=False,
                        agreement=state.current_offer,
                        times=times,
                        exceptions=exceptions,
                        broken=False,
                    )
            if resp.response == ResponseType.REJECT_OFFER:
                proposal = resp.outcome
                if (
                    not self.allow_offering_just_rejected_outcome
                    and proposal == state.current_offer
                ):
                    proposal = None
                if proposal is None:
                    if (
                        neg.capabilities.get("propose", True)
                        and self.end_negotiation_on_refusal_to_propose
                    ):
                        state.broken = True
                        return MechanismStepResult(
                            state, times=times, exceptions=exceptions
                        )
                    state.n_acceptances = 0
                else:
                    state.n_acceptances = 1 if self._offering_is_accepting else 0
                    if self._extra_callbacks:
                        for other in self.negotiators:
                            if other is neg:
                                continue
                            other.on_partner_proposal(
                                partner_id=neg.id, offer=proposal, state=self.state
                            )
                state.current_offer = proposal
                self._current_proposer = neg
                state.current_proposer = neg.id
                state.new_offers.append((neg.id, proposal))
                if self._last_checked_negotiator >= 0:
                    state.last_negotiator = self.negotiators[
                        self._last_checked_negotiator
                    ].name
                else:
                    state.last_negotiator = ""
                (
                    self._current_proposer_agent,
                    state.new_offerer_agents,
                ) = self._agent_info()

        # if action is not None:
        #     assert (
        #         not action
        #     ), f"Not all negotiator actions were used in this step: {action}"
        return MechanismStepResult(state, times=times, exceptions=exceptions)

    @property
    def full_trace(self) -> list[TraceElement]:
        """Returns the negotiation history as a list of relative_time/step/negotiator/offer tuples"""

        def response(state: SAOState):
            if state.agreement:
                return "agreement"
            if state.timedout:
                return "timedout"
            if state.ended:
                return "ended"
            if state.has_error:
                return "error"
            return "continuing"

        offers = []

        def get_acceptances(state: SAOState):
            neg = state.current_proposer
            n_acceptances = state.n_acceptances
            if self._offering_is_accepting:
                n_acceptances -= 1
            if neg is None:
                indices = []
            else:
                indx = self.negotiator_index(neg)
                n = self.nmi.n_negotiators
                if indx is None:
                    indices = []
                else:
                    indices = [
                        _ if _ < n else _ % n for _ in range(indx, n_acceptances + indx)
                    ]
            return [self.negotiator_ids[_] for _ in indices]

        for state in self._history:
            state: SAOState
            acceptances = get_acceptances(state)
            offers += [
                TraceElement(
                    state.time,
                    state.relative_time,
                    state.step,
                    n,
                    o,
                    {n: ResponseType.ACCEPT_OFFER for n in acceptances},
                    response(state),
                )
                for n, o in state.new_offers
            ]

        def not_equal(a, b):
            return any(x != y for x, y in zip(a, b))

        self._history: list[SAOState]
        # if the agreement does not appear as the last offer in the trace, add it.
        # this should not happen though!!
        if (
            self.agreement is not None
            and offers
            and not_equal(offers[-1].offer, self.agreement)
        ):
            acceptances = get_acceptances(self._history[-1])
            offers.append(
                TraceElement(
                    self._history[-1].time,
                    self._history[-1].relative_time,
                    self._history[-1].step,
                    self._history[-1].current_proposer,
                    self.agreement,
                    {n: ResponseType.ACCEPT_OFFER for n in acceptances},
                    response(self._history[-1]),
                )
            )

        return offers

    @property
    def extended_trace(self) -> list[tuple[int, str, Outcome]]:
        """Returns the negotiation history as a list of step/negotiator/offer tuples"""
        offers = []
        for state in self._history:
            state: SAOState
            offers += [(state.step, n, o) for n, o in state.new_offers]

        def not_equal(a, b):
            return any(x != y for x, y in zip(a, b))

        self._history: list[SAOState]  #
        # if the agreement does not appear as the last offer in the trace, add it.
        # this should not happen though!!
        if (
            self.agreement is not None
            and offers
            and not_equal(offers[-1][-1], self.agreement)
        ):
            offers.append(
                (
                    self._history[-1].step,
                    self._history[-1].current_proposer,
                    self.agreement,
                )
            )

        return offers

    @property
    def trace(self) -> list[tuple[str, Outcome]]:
        """Returns the negotiation history as a list of negotiator/offer tuples"""
        offers = []
        for state in self._history:
            offers += [(n, o) for n, o in state.new_offers]

        def not_equal(a, b):
            if isinstance(a, dict):
                a = a.values()
            if isinstance(b, dict):
                b = b.values()
            return any(x != y for x, y in zip(a, b))

        if (
            self.agreement is not None
            and offers
            and not_equal(offers[-1][-1], self.agreement)
        ):
            offers.append((self._history[-1].current_proposer, self.agreement))

        return offers

    def negotiator_offers(self, negotiator_id: str) -> list[Outcome]:
        """Returns the offers given by a negotiator (in order)"""
        return [o for n, o in self.trace if n == negotiator_id]

    def negotiator_full_trace(
        self, negotiator_id: str
    ) -> list[tuple[float, float, int, Outcome, str]]:
        """Returns the (time/relative-time/step/outcome/response) given by a negotiator (in order)"""
        return [
            (t, rt, s, o, a)
            for t, rt, s, n, o, _, a in self.full_trace
            if n == negotiator_id
        ]

    @property
    def offers(self) -> list[Outcome]:
        """Returns the negotiation history as a list of offers"""
        return [o for _, o in self.trace]

    @property
    def _step(self):
        """
        A private property used by the checkpoint system
        """
        return self._current_state.step

    def plot(
        self,
        plotting_negotiators: tuple[int, int] | tuple[str, str] = (0, 1),
        save_fig: bool = False,
        path: str | None = None,
        fig_name: str | None = None,
        ignore_none_offers: bool = True,
        with_lines: bool = True,
        show_agreement: bool = False,
        show_pareto_distance: bool = True,
        show_nash_distance: bool = True,
        show_kalai_distance: bool = True,
        show_max_welfare_distance: bool = True,
        show_max_relative_welfare_distance: bool = False,
        show_end_reason: bool = True,
        show_last_negotiator: bool = True,
        show_annotations: bool = False,
        show_reserved: bool = True,
        show_total_time=True,
        show_relative_time=True,
        show_n_steps=True,
        colors: list | None = None,
        markers: list[str] | None = None,
        colormap: str = DEFAULT_COLORMAP,
        ylimits: tuple[float, float] | None = None,
        common_legend: bool = True,
        xdim: str = "relative_time",
        only2d: bool = False,
        no2d: bool = False,
        fast: bool = False,
        simple_offers_view: bool = False,
        mark_offers_view: bool = True,
        mark_pareto_points: bool = True,
        mark_all_outcomes: bool = True,
        mark_nash_points: bool = True,
        mark_kalai_points: bool = True,
        mark_max_welfare_points: bool = True,
        **kwargs,
    ):
        from negmas.plots.util import plot_mechanism_run

        extra_annotation = (
            f"Last: {self._current_state.last_negotiator}"
            if show_last_negotiator
            else ""
        )
        return plot_mechanism_run(
            mechanism=self,
            negotiators=plotting_negotiators,
            save_fig=save_fig,
            path=path,
            fig_name=fig_name,
            ignore_none_offers=ignore_none_offers,
            with_lines=with_lines,
            only2d=only2d,
            show_agreement=show_agreement,
            show_pareto_distance=show_pareto_distance,
            show_nash_distance=show_nash_distance,
            show_kalai_distance=show_kalai_distance,
            show_max_welfare_distance=show_max_welfare_distance,
            show_max_relative_welfare_distance=show_max_relative_welfare_distance,
            show_end_reason=show_end_reason,
            show_annotations=show_annotations,
            show_reserved=show_reserved,
            colors=colors,
            markers=markers,
            colormap=colormap,
            ylimits=ylimits,
            common_legend=common_legend,
            extra_annotation=extra_annotation,
            xdim=xdim,
            colorizer=lambda _: 1.0,
            show_total_time=show_total_time,
            show_relative_time=show_relative_time,
            show_n_steps=show_n_steps,
            fast=fast,
            no2d=no2d,
            simple_offers_view=simple_offers_view,
            mark_offers_view=mark_offers_view,
            mark_pareto_points=mark_pareto_points,
            mark_all_outcomes=mark_all_outcomes,
            mark_nash_points=mark_nash_points,
            mark_kalai_points=mark_kalai_points,
            mark_max_welfare_points=mark_max_welfare_points,
            **kwargs,
        )


SAOProtocol = SAOMechanism
"""An alias for `SAOMechanism` object"""



from __future__ import annotations
from pytest import mark

from negmas import BinaryComparatorNegotiator, MappingUtilityFunction
from negmas.st import HillClimbingSTMechanism, VetoSTMechanism


@mark.parametrize(
    "n_negotiators,n_outcomes,n_steps",
    [(2, 10000, 300), (3, 50, 3), (2, 50, 3), (3, 5, 3)],
)
def test_can_work(n_negotiators, n_outcomes, n_steps):
    mechanism = VetoSTMechanism(outcomes=n_outcomes, n_steps=n_steps)
    ufuns = MappingUtilityFunction.generate_random(n_negotiators, outcomes=n_outcomes)
    for i in range(n_negotiators):
        mechanism.add(
            BinaryComparatorNegotiator(name=f"agent{i}"), preferences=ufuns[i]
        )
    assert mechanism.state.step == 0
    mechanism.step()
    assert mechanism.state.step == 1
    assert mechanism.state.current_offer is not None
    mechanism.run()
    assert mechanism.agreement is not None


@mark.parametrize("n_negotiators,n_outcomes", [(2, 10), (3, 50), (2, 50), (3, 5)])
def test_hill_climbing_mechanism(n_negotiators, n_outcomes):
    mechanism = HillClimbingSTMechanism(outcomes=n_outcomes, n_steps=3)
    ufuns = MappingUtilityFunction.generate_random(n_negotiators, outcomes=n_outcomes)
    for i in range(n_negotiators):
        mechanism.add(
            BinaryComparatorNegotiator(name=f"agent{i}"), preferences=ufuns[i]
        )
    assert mechanism.state.step == 0
    mechanism.step()
    assert mechanism.state.step == 1
    assert mechanism.state.current_offer is not None
    mechanism.run()
    assert mechanism.agreement is not None



from __future__ import annotations
from random import random
from typing import TYPE_CHECKING

from ..crisp_ufun import UtilityFunction
from ..mixins import StationaryMixin

if TYPE_CHECKING:
    from negmas.outcomes import Outcome

__all__ = ["RandomUtilityFunction"]


class RandomUtilityFunction(StationaryMixin, UtilityFunction):
    """A random utility function for a discrete outcome space"""

    def __init__(self, rng: tuple[float, float] = (0.0, 1.0), *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cache: dict[Outcome | None, float] = dict()
        self._scale = rng[1] - rng[0]
        self._offset = rng[0]

    def eval(self, offer: Outcome | None) -> float:
        v = self._cache.get(offer, None)
        if v is None:
            v = self._offset + self._scale * random()
            self._cache[offer] = v
        return v



from __future__ import annotations
import itertools
import math
from random import sample
from typing import TYPE_CHECKING, Callable

import numpy as np

from negmas.negotiators import Negotiator

from ..outcomes.issue_ops import sample_issues

if TYPE_CHECKING:
    from ..common import Value
    from ..outcomes.base_issue import Issue
    from ..outcomes.common import Outcome
    from ..preferences import Preferences

__all__ = [
    "EvaluatorNegotiator",
    "RealComparatorNegotiator",
    "BinaryComparatorNegotiator",
    "NLevelsComparatorNegotiator",
    "RankerNegotiator",
    "RankerWithWeightsNegotiator",
    "SorterNegotiator",
]


class EvaluatorNegotiator(Negotiator):
    """
    A negotiator that can be asked to evaluate outcomes using its internal ufun.

    Th change the way it evaluates outcomes, override `evaluate`.

    It has the `evaluate` capability
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.capabilities["evaluate"] = True

    def evaluate(self, outcome: Outcome) -> Value | None:
        if not self.ufun:
            return None
        return self.ufun(outcome)


class RealComparatorNegotiator(Negotiator):
    """
    A negotiator that can be asked to evaluate outcomes using its internal ufun.

    Th change the way it evaluates outcomes, override `compare_real`

    It has the `compare-real` capability
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.capabilities["compare-real"] = True
        self.capabilities["compare-binary"] = True

    def difference(self, first: Outcome, second: Outcome) -> float:
        """
        Compares two offers using the `ufun` returning the difference in their utility

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared

        Returns:
            "Value": An estimate of the differences between the two outcomes. It can be a real number between -1, 1
            or a probability distribution over the same range.
        """
        if not self.preferences:
            raise ValueError("Cannot compare outcomes. I have no preferences")
        return self.preferences.difference(first, second)  # type: ignore

    def is_better(self, first: Outcome | None, second: Outcome | None) -> bool | None:
        """
        Compares two offers using the `ufun` returning whether the first is better than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared

        Returns:
            True if utility(first) > utility(second) + epsilon
            None if |utility(first) - utility(second)| <= epsilon or the utun is not defined
            False if utility(first) < utility(second) - epsilon
        """
        if not self.preferences:
            return None
        return self.preferences.is_better(first, second)


class BinaryComparatorNegotiator(Negotiator):
    """
    A negotiator that can be asked to compare two outcomes using is_better. By default is just consults the ufun.

    To change that behavior, override `is_better`.

    It has the `compare-binary` capability.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.capabilities["compare-binary"] = True

    def is_better(
        self, first: Outcome | None, second: Outcome | None, epsilon: float = 1e-10
    ) -> bool | None:
        """
        Compares two offers using the `ufun` returning whether the first is better than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared
            epsilon: comparison threshold. If the utility difference within the range [-epsilon, epsilon] the two
                     outcomes are assumed to be compatible

        Returns:
            True if utility(first) > utility(second) + epsilon
            None if |utility(first) - utility(second)| <= epsilon or the utun is not defined
            False if utility(first) < utility(second) - epsilon
        """
        if not self.has_preferences:
            raise ValueError("Cannot compare outcomes without a ufun")
        return self._preferences.is_better(first, second)  # type: ignore


class NLevelsComparatorNegotiator(Negotiator):
    """
    A negotiator that can be asked to compare two outcomes using compare_nlevels which returns the strength of
    the difference between two outcomes as an integer from [-n, n] in the C compare sense.
    By default is just consults the ufun.

    To change that behavior, override `compare_nlevels`.

    It has the `compare-nlevels` capability.

    """

    def __init__(self, *args, thresholds: list[float] | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.thresholds = thresholds  # type: ignore I am not sure why
        self.capabilities["compare-nlevels"] = True
        self.capabilities["compare-binary"] = True
        self.__preferences_thresholds = None

    @classmethod
    def generate_thresholds(
        cls,
        n: int,
        ufun_min: float = 0.0,
        ufun_max: float = 1.0,
        scale: str | Callable[[float], float] | None = None,
    ) -> list[float]:
        """
        Generates thresholds for the n given levels assuming the ufun ranges and scale function

        Args:
            n: Number of scale levels (one side)
            ufun_min: minimum value of all utilities
            ufun_max: maximum value of all utilities
            scale: Scales the ufun values. Can be a callable or 'log', 'exp', 'linear'. If None, it is 'linear'

        """
        if isinstance(scale, str):
            scale = dict(  # type: ignore
                linear=lambda x: x, log=math.log, exp=math.exp
            ).get(scale, None)
        if scale is None:
            raise ValueError(f"Unknown scale function {scale}")
        thresholds = np.linspace(ufun_min, ufun_max, num=n + 2)[1:-1].tolist()
        thresholds = [scale(_) for _ in thresholds]  # type: ignore
        return thresholds

    @classmethod
    def equiprobable_thresholds(
        cls,
        n: int,
        preferences: Preferences,
        issues: list[Issue],
        n_samples: int = 1000,
    ) -> list[float]:
        """
        Generates thresholds for the n given levels where levels are equally likely approximately

        Args:
            n: Number of scale levels (one side)
            preferences: The utility function to use
            issues: The issues to generate the thresholds for
            n_samples: The number of samples to use during the process

        """
        samples = list(
            sample_issues(
                issues, n_samples, with_replacement=False, fail_if_not_enough=False
            )
        )
        n_samples = len(samples)
        diffs = []
        for i, first in enumerate(samples):
            n_diffs = min(10, n_samples - i - 1)
            for second in sample(samples[i + 1 :], k=n_diffs):
                diffs.append(abs(preferences.compare_real(first, second)))  # type: ignore
        diffs = np.array(diffs)
        _, edges = np.histogram(diffs, bins=n + 1)
        return edges[1:-1].tolist()

    @property
    def thresholds(self) -> list[float] | None:
        """Returns the internal thresholds and None if they do  not exist"""
        return self.__preferences_thresholds

    @thresholds.setter
    def thresholds(self, thresholds: list[float]) -> None:
        self.__preferences_thresholds = thresholds

    def compare_nlevels(
        self, first: Outcome, second: Outcome, n: int = 2
    ) -> int | None:
        """
        Compares two offers using the `ufun` returning an integer in [-n, n] (i.e. 2n+1 possible values) which defines
        which outcome is better and the strength of the difference (discretized using internal thresholds)

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared
            n: number of levels to use

        Returns:

            - None if either there is no ufun defined or the number of thresholds required cannot be satisfied
            - 0 iff |u(first) - u(second)| <= thresholds[0]
            - -i if  - thresholds[i-1] < u(first) - u(second) <= -thresholds[i]
            - +i if  thresholds[i-1] > u(first) - u(second) >= thresholds[i]

        Remarks:

            - thresholds is an internal array that can be set using `thresholds` property
            - thresholds[n] is assumed to equal infinity
            - n must be <= the length of the internal thresholds array. If n > that length, a ValueError will be raised.
              If n < the length of the internal thresholds array, the first n values of the array will be used
        """
        if not self.has_preferences:
            return None
        if self.thresholds is None:
            raise ValueError(
                f"Internal thresholds array is not set. Please set the threshold property with an array"
                f" of length >= {n}"
            )
        if len(self.thresholds) < n:
            raise ValueError(
                f"Internal thresholds array is only of length {len(self.thresholds)}. It cannot be used"
                f" to compare outcomes with {n} levels. len(self.thresholds) MUST be >= {n}"
            )
        if not self.ufun:
            raise ValueError("Unknown preferences")
        diff = float(self.ufun(first)) - float(self.ufun(second))
        sign = 1 if diff > 0.0 else -1
        for i, th in enumerate(self.thresholds):
            if diff < th:
                return sign * i
        return sign * n

    def is_better(
        self, first: Outcome | None, second: Outcome | None, epsilon: float = 1e-10
    ) -> bool | None:
        """
        Compares two offers using the `ufun` returning whether the first is better than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared
            epsilon: comparison threshold. If the utility difference within the range [-epsilon, epsilon] the two
                     outcomes are assumed to be compatible

        Returns:
            True if utility(first) > utility(second) + epsilon
            None if |utility(first) - utility(second)| <= epsilon or the utun is not defined
            False if utility(first) < utility(second) - epsilon
        """
        if self._preferences is None:
            return None
        return self._preferences.is_better(first, second)


class RankerWithWeightsNegotiator(Negotiator):
    """
    A negotiator that can be asked to rank outcomes returning rank and weight. By default is just consults the ufun.

    To change that behavior, override `rank_with_weights`.

    It has the `rank-weighted` capability.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.capabilities["rank-weighted"] = True
        self.capabilities["compare-binary"] = True

    def rank_with_weights(
        self, outcomes: list[Outcome] | None, descending=True
    ) -> list[tuple[int, float]]:
        """Ranks the given list of outcomes with weights. None stands for the null outcome. Outcomes of equal utility
        are ordered arbitrarily.

        Returns:

            - A list of tuples each with two values:
                - an integer giving the index in the input array (outcomes) of an outcome
                - the weight of that outcome
            - The list is sorted by weights descendingly

        """
        if not self.preferences:
            raise ValueError("Has no preferences. Cannot rank")
        return self.preferences.rank_with_weights(outcomes, descending)  # type: ignore

    def is_better(
        self, first: Outcome | None, second: Outcome | None, epsilon: float = 1e-10
    ) -> bool | None:
        """
        Compares two offers using the `ufun` returning whether the first is better than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared
            epsilon: comparison threshold. If the utility difference within the range [-epsilon, epsilon] the two
                     outcomes are assumed to be compatible

        Returns:
            True if utility(first) > utility(second) + epsilon
            None if |utility(first) - utility(second)| <= epsilon or the utun is not defined
            False if utility(first) < utility(second) - epsilon
        """
        if not self.has_preferences:
            return None
        return self._preferences.is_better(first, second)  # type: ignore


class RankerNegotiator(Negotiator):
    """
    A negotiator that can be asked to rank outcomes. By default is just consults the ufun.

    To change that behavior, override `rank`.

    It has the `rank` capability.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.capabilities["rank"] = True
        self.capabilities["compare-binary"] = True

    def rank(self, outcomes: list[Outcome | None], descending=True) -> list[int]:
        """Ranks the given list of outcomes. None stands for the null outcome.

        Returns:

            - A list of integers in the specified order of utility values of outcomes

        """
        if not self.preferences:
            raise ValueError("Unknown preferences. Cannot rank")
        return self.preferences.rank(outcomes, descending)  # type: ignore

    def is_better(
        self, first: Outcome | None, second: Outcome | None, epsilon: float = 1e-10
    ) -> bool | None:
        """
        Compares two offers using the `ufun` returning whether the first is better than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared
            epsilon: comparison threshold. If the utility difference within the range [-epsilon, epsilon] the two
                     outcomes are assumed to be compatible

        Returns:
            True if utility(first) > utility(second) + epsilon
            None if |utility(first) - utility(second)| <= epsilon or the utun is not defined
            False if utility(first) < utility(second) - epsilon
        """
        if self._preferences is None:
            return None
        return self._preferences.is_better(first, second)


class SorterNegotiator(Negotiator):
    """
    A negotiator that can be asked to rank outcomes returning rank without weight.
    By default is just consults the ufun.

    To change that behavior, override `sort`.

    It has the `sort` capability.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.capabilities["sort"] = True

    def sort(self, outcomes: list[Outcome | None], descending=True) -> None:
        """Ranks the given list of outcomes. None stands for the null outcome.

        Returns:

            - The outcomes are sorted IN PLACE.
            - There is no way to know if the ufun is not defined from the return value. Use `has_preferences` to check for
              the availability of the ufun

        """
        if not self.has_preferences:
            raise ValueError("Cannot sort outcomes. Unknown preferences")
        ranks = self._preferences.argrank(outcomes, descending)  # type: ignore
        ranks = list(itertools.chain(*tuple(ranks)))
        original = [_ for _ in outcomes]
        for i in range(len(outcomes)):
            if ranks[i] is None:
                outcomes[i] = None
                continue
            outcomes[i] = original[ranks[i]]  # type: ignore



"""
Implements single text negotiation mechanisms
"""

from __future__ import annotations
import time
from copy import deepcopy

from attrs import define, field

from negmas.common import MechanismAction, NegotiatorMechanismInterface
from negmas.negotiators.simple import BinaryComparatorNegotiator

from .mechanisms import Mechanism, MechanismState, MechanismStepResult
from .outcomes import Outcome

__all__ = ["VetoMTMechanism"]


@define
class MTState(MechanismState):
    """Defines extra values to keep in the mechanism state. This is accessible to all negotiators"""

    current_offers: list[Outcome | None] = field(factory=list)


class VetoMTMechanism(
    Mechanism[
        NegotiatorMechanismInterface,
        MTState,
        MechanismAction,
        BinaryComparatorNegotiator,
    ]
):
    """Base class for all muti-text mechanisms

    Args:
        *args: positional arguments to be passed to the base Mechanism
        **kwargs: keyword arguments to be passed to the base Mechanism
        n_texts: Number of current offers to keep
        initial_outcomes: initial set of outcomes outcome. If None, it will be selected by `next_outcome` which by default will choose it
                         randomly.
        initial_responses: Initial set of responses. One list per outcome in initial_outcomes.

    Remarks:

        - initial_responses is only of value when the number of negotiators that will join the negotiation is less then
          or equal to its length. By default it is not used for anything. Nevertheless, it is here because
          `next_outcome` may decide to use it with the `initial_outcome`


    """

    def __init__(
        self,
        *args,
        epsilon: float = 1e-6,
        n_texts: int = 10,
        initial_outcomes: list[Outcome | None] | None = None,
        initial_responses: tuple[tuple[bool]] | None = None,
        initial_state: MTState | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        state = self._current_state = initial_state if initial_state else MTState()

        self.add_requirements(
            {"compare-binary": True}
        )  # assert that all agents must have compare-binary capability
        state.current_offers = (
            initial_outcomes if initial_outcomes is not None else [None] * n_texts
        )
        """The current offer"""
        self.initial_outcomes = deepcopy(state.current_offers)
        """The initial offer"""
        self.last_responses = (
            [list(_) for _ in initial_responses]
            if initial_responses is not None
            else [None] * n_texts
        )
        """The responses of all negotiators for the last offer"""
        self.initial_responses = deepcopy(self.last_responses)
        """The initial set of responses. See the remarks of this class to understand its role."""
        self.epsilon = epsilon

    def next_outcome(self, outcome: Outcome | None) -> Outcome | None:
        """Generate the next outcome given some outcome.

        Args:
             outcome: The current outcome

        Returns:
            a new outcome or None to end the mechanism run

        """
        _ = outcome
        return self.random_outcomes(1)[0]

    def __call__(self, state: MTState, action=None) -> MechanismStepResult:
        """Single round of the protocol"""
        for i, current_offer in enumerate(state.current_offers):
            new_offer = self.next_outcome(current_offer)
            responses = []

            for neg in self.negotiators:
                strt = time.perf_counter()
                is_better = (
                    neg.is_better(new_offer, current_offer, epsilon=self.epsilon)
                    if action is None
                    else action.get(
                        neg.id,
                        neg.is_better(new_offer, current_offer, epsilon=self.epsilon),
                    )
                )
                responses.append(is_better is not False)
                if time.perf_counter() - strt > self.nmi.step_time_limit:
                    state.timedout = True
                    return MechanismStepResult(state)

            self.last_responses = responses

            if all(responses):
                self._current_state.current_offers[i] = new_offer

        return MechanismStepResult(state)

    def on_negotiation_end(self) -> None:
        """Used to pass the final offer for agreement between all negotiators"""
        self._agreement = self._current_state.current_offers  # type: ignore

        super().on_negotiation_end()



from __future__ import annotations
import random
from typing import Iterable

import numpy as np

from negmas.common import Distribution
from negmas.generics import gmap
from negmas.helpers import get_full_type_name
from negmas.helpers.prob import ScipyDistribution
from negmas.outcomes import Issue, Outcome
from negmas.outcomes.protocols import OutcomeSpace
from negmas.serialization import PYTHON_CLASS_IDENTIFIER, deserialize, serialize

from ..base import OutcomeUtilityMapping
from ..mixins import StationaryMixin
from ..prob_ufun import ProbUtilityFunction

__all__ = ["ProbMappingUtilityFunction"]


class ProbMappingUtilityFunction(StationaryMixin, ProbUtilityFunction):
    """
    Outcome mapping utility function.

    This is the simplest possible utility function and it just maps a set of `Outcome`s to a set of
    `Value`(s). It is only usable with single-issue negotiations. It can be constructed with wither a mapping
    (e.g. a dict) or a callable function.

    Args:
            mapping: Either a callable or a mapping from `Outcome` to `Value`.
            default: value returned for outcomes causing exception (e.g. invalid outcomes).
            name: name of the utility function. If None a random name will be generated.
            reserved_value: The reserved value (utility of not getting an agreement = utility(None) )

    Remarks:
        - If the mapping used failed on the outcome (for example because it is not a valid outcome), then the
        ``default`` value given to the constructor (which defaults to None) will be returned.

    """

    def __init__(
        self, mapping: OutcomeUtilityMapping, default=None, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.mapping = mapping
        self.default = default

    def to_dict(self):
        d = {PYTHON_CLASS_IDENTIFIER: get_full_type_name(type(self))}
        d.update(super().to_dict())
        return dict(**d, mapping=serialize(self.mapping), default=self.default)

    @classmethod
    def from_dict(cls, d):
        d.pop(PYTHON_CLASS_IDENTIFIER, None)
        d["mapping"] = deserialize(d["mapping"])
        return cls(**d)

    def eval(self, offer: Outcome | None) -> Distribution | float | None:
        # noinspection PyBroadException
        if offer is None:
            return self.reserved_value
        try:
            m = gmap(self.mapping, offer)
        except Exception:
            return self.default

        return m

    def xml(self, issues: list[Issue]) -> str:
        raise NotImplementedError("Cannot save ProbMappingUtilityFunction to xml")

    @classmethod
    def random(
        cls,
        outcome_space: OutcomeSpace,
        reserved_value=(0.0, 1.0),
        normalized=True,
        max_cardinality: int = 10000,
        type: str = "uniform",
    ):
        # todo: corrrect this for continuous outcome-spaces
        if not isinstance(reserved_value, Iterable):
            reserved_value = (reserved_value, reserved_value)
        os = outcome_space.to_largest_discrete(
            levels=10, max_cardinality=max_cardinality
        )
        mn, rng = 0.0, 1.0
        if not normalized:
            mn = 4 * random.random()
            rng = 4 * random.random()
        locs = np.random.rand(os.cardinality) * rng + mn
        scales = np.random.rand(os.cardinality) * rng + mn
        return cls(
            dict(
                zip(
                    os.enumerate(),
                    (
                        ScipyDistribution(type=type, loc=k, scale=s)
                        for k, s in zip(locs, scales)
                    ),
                )
            ),
            reserved_value=reserved_value[0]  # type: ignore
            + random.random() * (reserved_value[1] - reserved_value[0]),  # type: ignore
        )

    def __str__(self) -> str:
        return f"mapping: {self.mapping}\ndefault: {self.default}"
